"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const node_docker_api_1 = require("node-docker-api");
const node_fetch_1 = __importDefault(require("node-fetch"));
const container_1 = __importDefault(require("./container"));
const crypto_1 = require("crypto");
const ENGINE_NETWORK_NAME = 'engine';
const PREFIX = 'mesg_srv_';
const MAX_RETRY = 5;
class DockerContainer {
    constructor(opts = null, serviceEndpoint = ENGINE_NETWORK_NAME) {
        this.ipfsGateway = "http://ipfs.app.mesg.com:8080/ipfs";
        this._client = new node_docker_api_1.Docker(opts);
        this.serviceEndpoint = serviceEndpoint;
    }
    start(service, env, runnerHash) {
        return __awaiter(this, void 0, void 0, function* () {
            const labels = {
                'mesg.service': service.hash,
                'mesg.runner': runnerHash,
            };
            const image = yield this.build(service);
            let serviceNetwork = yield this.findNetwork(labels);
            if (!serviceNetwork)
                serviceNetwork = yield this._client.network.create({
                    name: PREFIX + service.hash,
                    labels
                });
            const engineNetwork = yield this.findNetwork({ 'mesg.engine': 'true' });
            if (!engineNetwork)
                throw new Error('engine network not found');
            for (const dep of service.dependencies || []) {
                const container = new container_1.default({
                    Args: dep.args,
                    Command: dep.command,
                    Env: dep.env,
                    Image: dep.image,
                    Labels: Object.assign(Object.assign({}, labels), { 'mesg.dependency': dep.key }),
                    HostConfig: {
                        Mounts: this.convertVolumes(service, dep.volumes, dep.volumesFrom, dep),
                        RestartPolicy: {
                            Name: 'on-failure',
                            MaximumRetryCount: MAX_RETRY
                        }
                    }
                }, PREFIX + service.hash + '_' + dep.key, this._client);
                container.addPorts(dep.ports);
                container.connectTo(serviceNetwork, [dep.key]);
                yield container.start();
            }
            const envObj = Object.assign(Object.assign(Object.assign({}, (service.configuration.env || []).reduce((prev, x) => (Object.assign(Object.assign({}, prev), { [x.split('=')[0]]: x.split('=')[1] })), {})), env), { MESG_ENDPOINT: `${this.serviceEndpoint}:50052` });
            const container = new container_1.default({
                Args: service.configuration.args,
                Command: service.configuration.command,
                Env: Object.keys(envObj).reduce((prev, x) => [...prev, `${x}=${envObj[x]}`], []),
                Image: image,
                Labels: Object.assign(Object.assign({}, labels), { 'mesg.dependency': 'service' }),
                HostConfig: {
                    Mounts: this.convertVolumes(service, service.configuration.volumes, service.configuration.volumesFrom),
                    RestartPolicy: {
                        Name: 'on-failure',
                        MaximumRetryCount: MAX_RETRY
                    }
                }
            }, PREFIX + service.hash, this._client);
            container.addPorts(service.configuration.ports);
            container.connectTo(serviceNetwork, ['service']);
            container.connectTo(engineNetwork, ['engine']);
            yield container.start();
            return true;
        });
    }
    stop(runnerHash, serviceHash) {
        return __awaiter(this, void 0, void 0, function* () {
            const labels = {
                'mesg.service': serviceHash,
                'mesg.runner': runnerHash,
            };
            const containers = yield container_1.default.findAll(this._client, labels);
            for (const container of containers) {
                yield container.stop();
                yield container.delete();
            }
            const serviceNetwork = yield this.findNetwork(labels);
            if (serviceNetwork)
                yield serviceNetwork.remove();
        });
    }
    findNetwork(labels) {
        return __awaiter(this, void 0, void 0, function* () {
            const allLabels = (data) => Object.keys(labels)
                .reduce((prev, x) => prev && data[x] === labels[x], true);
            const list = yield this._client.network.list();
            const items = list.filter(x => allLabels(x.data['Labels'] || {}));
            if (items.length > 1)
                throw new Error(`more than one network found`);
            return items[0];
        });
    }
    build(service) {
        return __awaiter(this, void 0, void 0, function* () {
            const streamToPromise = (stream) => {
                let res = '';
                return new Promise((resolve, reject) => stream
                    .on('error', reject)
                    .on('data', (x) => res += x.toString())
                    .on('end', () => {
                    const logs = res.split('\n').map(x => {
                        try {
                            return JSON.parse(x);
                        }
                        catch (e) {
                            return null;
                        }
                    }).filter(x => x);
                    const error = logs.find((x) => x.error);
                    if (error)
                        return reject(error.error);
                    return resolve();
                }));
            };
            const resp = yield node_fetch_1.default(`${this.ipfsGateway}/${service.source}`);
            const tag = `mesg:${service.hash}`;
            const image = yield this._client.image.build(resp.body, {
                t: tag,
                q: true
            });
            yield streamToPromise(image);
            return tag;
        });
    }
    convertVolumes(service, volumes, volumesFrom, dependency) {
        const res = [];
        for (const volume of volumes || []) {
            res.push({
                Source: this.volumeName(service, volume, dependency),
                Target: volume,
                Type: 'volume'
            });
        }
        // Add volumes from
        for (const dep of volumesFrom || []) {
            const dependency = (service.dependencies || []).find(x => x.key === dep);
            if (!dependency)
                throw new Error(`dependency "${dependency}" missing`);
            for (const volume of dependency.volumes) {
                res.push({
                    Source: this.volumeName(service, volume, dependency),
                    Target: volume,
                    Type: 'volume'
                });
            }
        }
        return res;
    }
    volumeName(service, vol, dependency) {
        return crypto_1.createHash('sha256')
            .update(JSON.stringify([
            service.hash,
            dependency ? dependency.key : null,
            vol
        ]))
            .digest('hex');
    }
}
exports.default = DockerContainer;
